'use strict';

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = build;

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var $registeredMacros = Symbol('registeredMacros');
var $macroState = Symbol('macroState');

/**
 * # Babel Macros
 */
function build(babel) {
  if (!(babel instanceof Object)) {
    throw new TypeError('Value of argument "babel" violates contract.\n\nExpected:\nObject\n\nGot:\n' + _inspect(babel));
  }

  var t = babel.types;
  var traverse = babel.traverse;

  /**
   * A list of builtin macros.
   */

  var builtin = Object.create(null);

  var Macro = function () {
    function Macro(_ref4) {
      var name = _ref4.name;
      var macroBody = _ref4.macroBody;
      var scope = _ref4.scope;
      var state = _ref4.state;

      _classCallCheck(this, Macro);

      this.name = name;
      this.macroBody = macroBody;
      this.scope = scope;
      this.state = state;
    }

    _createClass(Macro, [{
      key: 'run',
      value: function run(path, scope, state) {
        if (!this.macro) {
          this.macro = compileMacro(this.name, this.macroBody, this.scope, this.state);
        }
        return this.macro(path, scope, state);
      }
    }]);

    return Macro;
  }();

  builtin.DEFINE_MACRO = function defineMacro(path, scope, state) {
    var node = path.node;

    var id = node.arguments[0];
    if (!(id && t.isIdentifier(id))) {
      // @todo add test
      throw new Error('First argument to DEFINE_MACRO must be an identifier.');
    }
    var name = id.name;
    var macroBody = node.arguments[1];
    var subScope = path.get('arguments')[1].scope;

    scope[$registeredMacros] = scope[$registeredMacros] || {};
    scope[$registeredMacros][name] = new Macro({ name: name, macroBody: macroBody, scope: subScope, state: state });
    traverse(macroBody, visitors, subScope, state);
    path.remove();
  };

  function compileMacro(name, node, scope, state) {
    function _ref2(_id2) {
      if (!(typeof _id2 === 'function')) {
        throw new TypeError('Function "compileMacro" return value violates contract.\n\nExpected:\nFunction\n\nGot:\n' + _inspect(_id2));
      }

      return _id2;
    }

    if (!(typeof name === 'string')) {
      throw new TypeError('Value of argument "name" violates contract.\n\nExpected:\nstring\n\nGot:\n' + _inspect(name));
    }

    if (!(node instanceof Object)) {
      throw new TypeError('Value of argument "node" violates contract.\n\nExpected:\nObject\n\nGot:\n' + _inspect(node));
    }

    if (!(scope instanceof Object)) {
      throw new TypeError('Value of argument "scope" violates contract.\n\nExpected:\nObject\n\nGot:\n' + _inspect(scope));
    }

    if (!(state instanceof Object)) {
      throw new TypeError('Value of argument "state" violates contract.\n\nExpected:\nObject\n\nGot:\n' + _inspect(state));
    }

    var paramNames = node.params.map(function (param) {
      return param.name;
    });
    var paramReferenceCounts = {};
    var references = Object.create(null);
    if (t.isFunction(node)) {
      traverse(node, visitors, scope, state);
      traverse(node, {
        enter: function enter(subPath) {
          var child = subPath.node;
          var parent = subPath.parent;

          if (subPath.isVariableDeclarator() || subPath.isFunctionDeclaration()) {
            references[child.id.name] = true;
          } else if (subPath.isIdentifier() && (!t.isFunction(parent) || parent.type === "ArrowFunctionExpression" && parent.body === child) && (!t.isMemberExpression(parent) || parent.object === child) && ~paramNames.indexOf(child.name)) {
            paramReferenceCounts[child.name] = paramReferenceCounts[child.name] || 0;
            paramReferenceCounts[child.name]++;
          }
        }
      }, scope);
    } else {
      // @todo add test
      throw new Error('Second argument to DEFINE_MACRO must be a FunctionExpression or ArrowFunctionExpression.');
    }
    return _ref2(function (path, scope, state) {
      var cloned = _lodash2.default.cloneDeep(node);

      var _cloned$params$reduce = cloned.params.reduce(function (_ref7, id, index) {
        var _ref8 = _slicedToArray(_ref7, 2);

        var params = _ref8[0];
        var seen = _ref8[1];

        params[id.name] = {
          id: id,
          replacement: path.node.arguments[index] || t.identifier('undefined'),
          reference: null
        };
        seen[id.name] = false;
        return [params, seen];
      }, [{}, {}]);

      var _cloned$params$reduce2 = _slicedToArray(_cloned$params$reduce, 2);

      var params = _cloned$params$reduce2[0];
      var seen = _cloned$params$reduce2[1];

      var blockStack = [];
      var loopStack = [];
      var returnStatements = [];
      var hasEarlyReturn = false;

      traverse(cloned, {
        enter: function enter(subPath) {
          var child = subPath.node;
          var parent = subPath.parent;

          if (child.type === 'Identifier' && (parent.type !== "MemberExpression" || parent.object === child || parent.computed && parent.property === child)) {
            if (params[child.name]) {
              var param = params[child.name];
              if (param.replacement.type === 'Identifier' || param.replacement.type === 'Literal' || paramReferenceCounts[child.name] === 1 && param.replacement.type === 'MemberExpression') {
                subPath.replaceWith(param.replacement);
                seen[child.name] = param.replacement;
              } else {
                if (!seen[child.name]) {
                  seen[child.name] = scope.generateUidIdentifier(child.name);

                  getParentBlock(path).insertBefore([t.variableDeclaration('const', [t.variableDeclarator(seen[child.name], param.replacement)])]);
                }
                subPath.replaceWith(seen[child.name]);
              }
            } else if (references[child.name]) {
              if (!seen[child.name]) {
                seen[child.name] = scope.generateUidIdentifier(child.name);
              }
              subPath.replaceWith(seen[child.name]);
            }
          } else if (subPath.isReturnStatement()) {
            if (blockStack.length > 0) {
              hasEarlyReturn = true;
            }
            returnStatements.push([subPath, child]);
          } else if (subPath.isLoop()) {
            loopStack.push(subPath);
          } else if (subPath.isFunction()) {
            subPath.skip();
          } else if (subPath.isScope()) {
            blockStack.push(subPath);
          }
        },
        exit: function exit(path) {
          if (path.isLoop()) {
            loopStack.pop();
          } else if (path.isScope()) {
            blockStack.pop();
          }
        }
      }, scope);

      if (t.isStatement(cloned.body)) {
        (function () {
          var uid = scope.generateUidIdentifier(camelCase(name));
          var labelUid = scope.generateUidIdentifier('_' + name.toUpperCase());
          var parentBlock = getParentBlock(path);
          parentBlock.insertBefore([t.variableDeclaration('let', [t.variableDeclarator(uid)])]);
          returnStatements.forEach(function (_ref5) {
            var _ref6 = _slicedToArray(_ref5, 2);

            var path = _ref6[0];
            var child = _ref6[1];

            var isLast = child === cloned.body.body[cloned.body.body.length - 1];
            path.replaceWith(t.expressionStatement(t.assignmentExpression('=', uid, child.argument)));

            if (hasEarlyReturn && !isLast) {
              path.insertAfter(t.breakStatement(labelUid));
            }
          });

          if (hasEarlyReturn) {
            parentBlock.insertBefore(t.labeledStatement(labelUid, cloned.body));
          } else {
            parentBlock.insertBefore(cloned.body.body);
          }
          path.replaceWith(uid);
        })();
      } else {
        path.replaceWith(cloned.body);
      }
    });
  }

  function runMacro(path, macro, scope, state) {
    if (typeof macro === 'function') {
      macro(path, scope, state);
    } else {
      macro.run(path, scope, state);
    }
  }

  function getParentBlock(path) {
    while (path.parentPath.type !== 'Program' && path.parentPath && !path.parentPath.isStatementOrBlock()) {
      path = path.parentPath;
    }
    return path;
  }

  function getParentScope(path) {
    while (path.parentPath.type !== 'Program' && path.parentPath && !path.parentPath.isStatementOrBlock()) {
      path = path.parentPath;
    }
    return path.scope;
  }

  function camelCase(input) {
    return input.toLowerCase().replace(/_(.)/g, function (match, char) {
      return char.toUpperCase();
    });
  }

  function getMacro(node, scope, state) {
    function _ref3(_id3) {
      if (!(_id3 instanceof Macro || typeof _id3 === 'function' || _id3 == null)) {
        throw new TypeError('Function "getMacro" return value violates contract.\n\nExpected:\nMacro | Function | void\n\nGot:\n' + _inspect(_id3));
      }

      return _id3;
    }

    if (!(node instanceof Object)) {
      throw new TypeError('Value of argument "node" violates contract.\n\nExpected:\nObject\n\nGot:\n' + _inspect(node));
    }

    if (!(scope instanceof Object)) {
      throw new TypeError('Value of argument "scope" violates contract.\n\nExpected:\nObject\n\nGot:\n' + _inspect(scope));
    }

    if (!(state instanceof Object)) {
      throw new TypeError('Value of argument "state" violates contract.\n\nExpected:\nObject\n\nGot:\n' + _inspect(state));
    }

    if (node.type === 'CallExpression') {
      return _ref3(getMacro(node.callee, scope, state));
    } else if (t.isIdentifier(node)) {
      if (state[$macroState].macrosDefined) {
        while (scope) {
          if (scope[$registeredMacros] && scope[$registeredMacros][node.name]) {
            return _ref3(scope[$registeredMacros][node.name]);
          }
          scope = scope.parent;
        }
      }
      if (builtin[node.name]) {
        return _ref3(builtin[node.name]);
      }
    } else if (t.isMemberExpression(node) && !node.computed && t.isIdentifier(node.property)) {
      return _ref3(getMacro(node.property, scope, state));
    }
  }

  var visitors = {
    CallExpression: {
      enter: function enter(path, state) {
        var node = path.node;
        if (state[$macroState].macrosDefined) {
          if (node._processedByMacro) {
            return;
          }
        }
        if (t.isMemberExpression(node.callee)) {
          // TODO temp locked for refactoring. need another idea for chaining more than 2 macros
          if (!node.callee.computed && getMacro(node.callee.object, path.scope, state) && getMacro(node.callee.property, path.scope, state)) {
            var head = node.callee.object;
            var tailId = node.callee.property;
            node.arguments.unshift(head); //TODO - so slow. maybe is exponent slow ?
            var macro = getMacro(tailId, path.scope, state);
            if (macro) {
              runMacro(path, macro, path.scope, state);
            }
          }
        } else {
          var macro = getMacro(node.callee, path.scope, state);
          if (macro) {
            runMacro(path, macro, path.scope, state);
          }
        }

        if (state[$macroState].macrosDefined) {
          node._processedByMacro = true;
        }
      }
    },
    Program: {
      enter: function enter(path, state) {
        state[$macroState] = {
          macrosDefined: false
        };
      },
      exit: function exit(path, state) {
        if (!state[$macroState].macrosDefined) {
          state[$macroState].macrosDefined = true;
          path.traverse(visitors, state);
        }
      }
    }
  };

  /**
   * Export the transformer.
   */
  return { visitor: visitors };
}

function _inspect(input) {
  if (input === null) {
    return 'null';
  } else if (input === undefined) {
    return 'void';
  } else if (typeof input === 'string' || typeof input === 'number' || typeof input === 'boolean') {
    return typeof input === 'undefined' ? 'undefined' : _typeof(input);
  } else if (Array.isArray(input)) {
    if (input.length > 0) {
      var first = _inspect(input[0]);

      if (input.every(function (item) {
        return _inspect(item) === first;
      })) {
        return first.trim() + '[]';
      } else {
        return '[' + input.map(_inspect).join(', ') + ']';
      }
    } else {
      return 'Array';
    }
  } else {
    var keys = Object.keys(input);

    if (!keys.length) {
      if (input.constructor && input.constructor.name && input.constructor.name !== 'Object') {
        return input.constructor.name;
      } else {
        return 'Object';
      }
    }

    var entries = keys.map(function (key) {
      return (/^([A-Z_$][A-Z0-9_$]*)$/i.test(key) ? key : JSON.stringify(key)) + ': ' + _inspect(input[key]) + ';';
    }).join('\n  ');

    if (input.constructor && input.constructor.name && input.constructor.name !== 'Object') {
      return input.constructor.name + ' {\n  ' + entries + '\n}';
    } else {
      return '{ ' + entries + '\n}';
    }
  }
}